# =====================================================================
# xmlrpc-client.awk: client XML-RPC AWK functions for CSA.
#
# Copyright (c) 2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: xmlrpc-client.awk,v 1.1 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# string xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appare anywhere in adocument, not even as character references, such
# as &#31;. The only other possibility it to either remove them or turn
# them into something sensible. I took this second approach, whereby
# \001 (that is rc(1) list concatenation character) is replaced with a
# single blank. This is experimental, and may change in future versions
# of CSA.
# =====================================================================

function xmlencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/"/, "\\&quot;", string)		# double quote
   gsub(/\001/, " ", string)			# rc(1)'s ^A

   return string
}

# =====================================================================
# string xmldecode(string s)
#
# Perform the inverse of xmlencode(), with the exception that the \001
# hack cannot be reverted, not being it distinquishable from all other
# blanks.
# =====================================================================

function xmldecode(string) {

   gsub(/&quot;/, "\"", string)			# double quote
   gsub(/&lt;/, "<", string)			# Open tag
   gsub(/&amp;/, "\\&", string)			# Ampersand at last.

   return string
}

# =====================================================================
# nil httpprint(string s)
#
# Special "print" function for HTTP messages, that does buffering and
# HTTP content-length calculation if necessary. Uses the HTTPBODY[]
# global array, so that its contents are preserved across multiple
# invocation of this function. The response size is stored in the
# HTTPLEN special var, for the same reason. HTTPBODY[0] will contain the
# no. of elements in the array.
# =====================================================================

function httpprint(s) {

  HTTPBODY[++HTTPBODY[0]] = s
  HTTPLEN += length(s)
}

# =====================================================================
# int envtoxml([string template, string tmpfile, string method,\
#		string eval, string table, boolean trim, string tago,\
#		string tagc, string cmd, string url, string pyxfile,
#		string errfile])
#
# Take environment variables and use them to replace special tags in
# an XML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
#
# The lightweight Boa Web Server does not set the Content-Length header
# field in HTTP replies. This is correct, IMHO, as computing such
# header may be prohibitively expensive in some cases, because we need
# to buffer the whole response either into memory or to disk, before
# such header can be computed. However, most XML-RPC servers insist
# on pretending thast such header be present, and fail otherwise.
# To make such servers happy I added to CSA the ability to compute
# such header when needed.
# =====================================================================

function envtoxml(tpl,tmp,method,eval,tbl,\
	trim,tago,tagc,cmd,url,pyx,err,				 \
	ltago,ltagc,retago,retagc,end,tlen,tag,enc,a,i,in1,in2,j,\
	cmt,path,host,port,prt,rlen) {

  i=j=cmt=0

  # Check args and set defaults.
  if (tpl !~ /^\//) tpl = "/dev/null"	      # defaults if not full-path
  if (tmp !~ /^\//) tmp = "/dev/null"
  if (pyx !~ /^\//) pyx = "/dev/null"
  if (err !~ /^\//) err = "/dev/null"
  if (cmd == "") cmd = "nc"
  if (method == "") method = "system.listMethods"
  if (!sub(/^http[s]?:\/\//,"",url)) url = "localhost/RPC2"
  path = host = port = url
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return EINVAL # start- and end-tags must be different

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = escreg(tago)
  retagc = escreg(tagc)

  retago = retago "[-A-Za-z0-9_/.:]+" retagc

  # Load the extra table data if requested. This data is also loaded
  # in ENVIRON[], so it can be used to override data in there. The
  # relevant values will be assigned variables csat1, csat2, etc., which
  # are CSA-reserved names (see the docs), and the "t" suffix stands
  # for "table".

  if (tbl != "") {
     while (getline record < tbl > 0) {
	# Skip table header.
	if (record !~ /^\001/) j=split(record,a,"\t")
     }

     close(tbl)

     # Append to ENVIRON, in deterministic order.
     for (i=1; i<=j; i++) ENVIRON["csat" i] = a[j]
     delete a
     i = 0
  }

  for (j in ENVIRON) {

      # Always skip rc(1) functions, confidential data
      # and other CSA-specific stuff.

      if (j ~ /(^fn_|REMOTE_USER|PGPPASS$|^CSA_AUTH_PW|^csa0)/) {
	 ENVIRON[j] = "<!-- removed -->"
	 continue
      }

      # Call local RPC output hook on variable.
      ENVIRON[j] = outvar(j,ENVIRON[j])

      # Provide also the associated XML-encoded values.
      ENVIRON["XML_" j] = xmlencode(ENVIRON[j])
  }

  if (eval != "") split(eval,a,",")

  # Clear these variables before the very first call to httpprint().
  delete HTTPBODY; HTTPLEN=0

  # Deal with HTTP headers

  sub(/\/.*$/,"",host); if (host == "") host = "localhost"

  sub(/.*:/,"",port); sub(/\/.*/,"",port)
  if (port == "" || port !~ /^[0-9]+$/) {
     if (cmd ~ /ssl/) port = 443
     else port = 80
  }

  if (cmd ~ /ssl/) cmd = cmd " " host ":" port		# openssl
  else cmd = cmd " " host " " port			# netcat

  sub(/^[^\/]+/,"",path); if (path !~ /^\//) path = "/RPC2"

  # Build request msg header.

  HTTPHDR[++HTTPHDR[0]] = "POST " path " HTTP/1.0"
  HTTPHDR[++HTTPHDR[0]] = "Host: " host
  HTTPHDR[++HTTPHDR[0]] = "User-Agent: CSA/" ENVIRON["CSA_VERSION"]
  HTTPHDR[++HTTPHDR[0]] = "Connection: close"
  HTTPHDR[++HTTPHDR[0]] = "Content-Type: text/xml"
  HTTPHDR[++HTTPHDR[0]] = "Content-Length: "

  # Build request msg body.

  httpprint("<?xml version=\"1.0\"?><methodCall>")
  httpprint("<methodName>" method "</methodName><params>")

  # Process the input template file.

  while (getline in1 < tpl > 0) {

     # Deal with comments.

     # Detect continuation/end of multiline comment.
     if (cmt) {
	if (sub(/^[^<]*-->/,"",in1)) cmt=0
	else continue
     }

     # Remove inline comments.
     gsub(/<!--[^<]*-->/,"",in1)

     # Detect begin of multiline comment.
     if (sub(/^<!--[^<]*$/,"",in1)) { cmt=1; continue }

     # Restore trailing newline dropped by 'getline',
     # where appropriate:

     # detect begin of string|base64.
     if (in1 ~ /<(stri|b)[^>]+>/) nl="\n"

     # detect end of string|base64|value.
     if (in1 ~ /<\/(stri|b|v)[^>]+>/) nl=""

     in1 = in1 nl

     do {
        match(in1,retago)
        if (RSTART) {
	   end = index(in1, tagc)
	   tlen = end - RSTART - ltago
	   httpprint(substr(in1, 1, RSTART - 1))
	   tag = substr(in1, RSTART + ltago, tlen)

	   # Now handle encoding flags. Only base64 and plain strings
	   # are supported.
	   if (sub(/:m$/,NULL,tag)) {		# Base64.
	      enc = "base64"
	   }

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 # Print tag literally.
		 httpprint("$[" tag "]")
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }

		 if (enc == "base64") {		# Base64.
		    CMD = "mimencode -p " tag
		    # No need to preserve newlines in base64.
		    while (CMD | getline in2 > 0) httpprint(in2)
		    close(CMD)
		 }
		 else {
		    # generic text, perform XML escaping.
		    while (getline in2 < tag > 0)
			httpprint(xmlencode(in2"\n"))
		    close(tag)
		 }
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = strip(ENVIRON[tag])
	      httpprint(ENVIRON[tag])
	   }

	   in1 = substr(in1, end + ltagc)
	}
	else {
	   httpprint(in1)
	   in1 = NULL		# Not to loop forever.
	}
     }  while (length(in1))
  }

  close(tpl)

  httpprint("</params></methodCall>")

  HTTPHDR[HTTPHDR[0]] = HTTPHDR[HTTPHDR[0]] HTTPLEN "\n"

  # Set new command.

  if (ENVIRON["CSA_DEBUG"] == 1)
     cmd = "tee --append $CSA_ROOT/var/debug.log | " cmd
  CMD = cmd " > " tmp

  i=0; while (++i <= HTTPHDR[0]) print HTTPHDR[i] | CMD
  i=0; while (++i <= HTTPBODY[0]) printf("%s", HTTPBODY[i]) | CMD
  printf("\n") | CMD
  close(CMD)

  # Add a known DTD to validate the server reply, or discard the
  # message altogether if it already contains a DTD reference.

  CMD = ""
  if (ENVIRON["CSA_DEBUG"] == 1)
     CMD = "tee --append $CSA_ROOT/var/debug.log | " 
  CMD = CMD "rxpyx -o " pyx " 2>" err

  prt=0
  while (getline < tmp > 0) {
    if (!prt && $0 ~ /^HTTP\/1\.[01] /) {
       if ($0 !~ /^HTTP\/1\.[01] 200 /) {
	  # Exit on errors from the remote HTTP server.
	  print > err
	  break
       }
    }
    if ($0 ~ /^[\r]?$/) {prt=1; continue}    # detect end-of-header
    if ($0 ~ /<!DOCTYPE /) break	     # Inline DTD not allowed
    match($0,/^<\?xml [^>]*\?>/)	     # Set RLENGTH
    if ($0 ~ /<methodResponse>/) {
       printf("%s", substr($0,1,RLENGTH)) | CMD

       # Insert known DTD before validation.
       printf("<!DOCTYPE methodResponse SYSTEM \"%s/lib/xml-rpc.dtd\"",\
		CSAINST) | CMD
       printf(" [<!ENTITY %% XMLRPC.version2 \"INCLUDE\">]>") | CMD
       printf("%s\n", substr($0,RLENGTH+1)) | CMD
       continue
    }

    if (prt) print | CMD
  }

  # Test whether the cmd is open before closing it, not to overwrite
  # HTTP server errors that were written to the same error file.
  if (prt) close(CMD)

  # Free memory.
  delete HTTPHDR; delete HTTPBODY; HTTPLEN=0
}

# =====================================================================
# End of xmlrpc-client.awk
# =====================================================================
